#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import sys
import inspect
import logging

import skriba.messenger
from datetime import datetime

CURRENT_FUNCTION = 0


class LoggingFormatter(logging.Formatter):
    start_msg = "[ {time} ] ".format(time=skriba.messenger.SystemMessage.magenta_dim("%(asctime)s"))
    middle_msg = "{level}".format(level="%(levelname)-8s")
    execution_msg = "{name} [ {filename} ]: {exec_info}: ".format(
        name="%(name)-10s",
        filename=skriba.messenger.SystemMessage.grey("%(filename)-21s"),
        exec_info=skriba.messenger.SystemMessage.blue_dim("%(callchain)-30s"))

    message = "%(message)s"

    FORMATS = {
        logging.DEBUG: start_msg + skriba.messenger.SystemMessage.green(middle_msg) + execution_msg + message,
        logging.INFO: start_msg + skriba.messenger.SystemMessage.blue(middle_msg) + execution_msg + message,
        logging.WARNING: start_msg + skriba.messenger.SystemMessage.yellow(middle_msg) + execution_msg + message,
        logging.ERROR: start_msg + skriba.messenger.SystemMessage.red(middle_msg) + execution_msg + message,
        logging.CRITICAL: start_msg + skriba.messenger.SystemMessage.red_highlight(middle_msg) + execution_msg + message
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_logger(logger_name=None):
    from dask.distributed import get_worker

    function = inspect.stack()[0].function
    module_path = inspect.stack()[0].filename
    module = os.path.basename(module_path).strip(".py")
    lineno = inspect.stack()[0].lineno

    #extra = {'callchain': f"{module}.{function}:{lineno}"}
    extra = {'callchain': "%(module)s.%(fileName)s:%(lineno)"}

    if logger_name is None:
        logger_name = "LOGGER"

    try:
        worker = get_worker()

    except ValueError:
        # Scheduler processes
        logger_dict = logging.Logger.manager.loggerDict
        if logger_name in logger_dict:
            logger = logging.getLogger(logger_name)
        else:
            # If main logger is not started using client function it defaults to printing to term.
            logger = logging.getLogger(logger_name)
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(LoggingFormatter())
            logger.addHandler(handler)
            logger.setLevel(logging.getLevelName('INFO'))

        logger = logging.LoggerAdapter(logger, extra)
        return logger

    try:
        logger = worker.plugins['worker_logger'].logger
        logger = logging.LoggerAdapter(logger, extra)
        return logger

    except Exception as error:
        print("Could not load worker logger: {}".format(error))
        print(worker.plugins.keys())

        logger = logging.LoggerAdapter(logger, extra)
        return logging.getLogger()


def setup_logger(
        logger_name=None,
        log_to_term=False,
        log_to_file=True,
        log_file='LOGGER',
        log_level='INFO',
):
    """To set up as many loggers as you want"""
    if logger_name is None:
        logger_name = "LOGGER"

    function = inspect.stack()[0].function
    module_path = inspect.stack()[0].filename
    module = os.path.basename(module_path).strip(".py")
    lineno = inspect.stack()[0].lineno

    #extra = {'callchain': f"{module}.{function}:{lineno}"}
    extra = {'callchain': "%(module)s.%(fileName)s:%(lineno)"}

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.getLevelName(log_level))

    logger.handlers.clear()

    if log_to_term:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(LoggingFormatter())
        logger.addHandler(handler)

    if log_to_file:
        log_file = log_file + datetime.today().strftime('%Y%m%d_%H%M%S') + '.log'
        handler = logging.FileHandler(log_file)
        handler.setFormatter(LoggingFormatter())
        logger.addHandler(handler)

    logger = logging.LoggerAdapter(logger, extra)
    return logger


def get_worker_logger_name(logger_name=None):
    from dask.distributed import get_worker
    if logger_name is None:
        logger_name = "LOGGER"

    return logger_name + '_' + str(get_worker().id)


def setup_worker_logger(
        logger_name,
        log_to_term,
        log_to_file,
        log_file,
        log_level,
        worker_id
):
    parallel_logger_name = logger_name + '_' + str(worker_id)

    function = inspect.stack()[0].function
    module_path = inspect.stack()[0].filename
    module = os.path.basename(module_path).strip(".py")
    lineno = inspect.stack()[0].lineno

    #extra = {'callchain': f"{module}.{function}:{lineno}"}
    extra = {'callchain': "%(module)s.%(fileName)s:%(lineno)"}

    logger = logging.getLogger(parallel_logger_name)
    logger.setLevel(logging.getLevelName(log_level))

    if log_to_term:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(LoggingFormatter())
        logger.addHandler(handler)

    if log_to_file:
        log_file = log_file + '_' + str(worker_id) + '_' + datetime.today().strftime('%Y%m%d_%H%M%S') + '.log'
        handler = logging.FileHandler(log_file)
        handler.setFormatter(LoggingFormatter())
        logger.addHandler(handler)

    logger = logging.LoggerAdapter(logger, extra)
    return logger



import inspect
import time

from colorama import Fore, Back, Style

CURRENT_FUNCTION = 0
PREVIOUS_FUNCTION = 1


class SystemMessage:
    def __init__(self):
        pass

    @staticmethod
    def error(message, **kwargs):
        function_name = inspect.stack()[PREVIOUS_FUNCTION].function
        print("[ {time} ] [{level}] [{function}]: {message}".format(
            function=function_name,
            message=message,
            level=SystemMessage.red("error"),
            time=time.strftime("%x %H:%M:%S %Z", time.localtime()),
            **kwargs
        ))

    @staticmethod
    def info(message, **kwargs):
        function_name = inspect.stack()[PREVIOUS_FUNCTION].function
        print("[ {time} ] [{level}] [{function}]: {message}".format(
            function=function_name,
            message=message,
            level=SystemMessage.blue("info"),
            time=time.strftime("%x %H:%M:%S %Z", time.localtime()),
            **kwargs
        ))

    @staticmethod
    def debug(message, **kwargs):
        function_name = inspect.stack()[PREVIOUS_FUNCTION].function
        print("[ {time} ] [{level}] [{function}]: {message}".format(
            function=function_name,
            message=message,
            level=SystemMessage.green("debug"),
            time=time.strftime("%x %H:%M:%S %Z", time.localtime()),
            **kwargs
        ))

    @staticmethod
    def warning(message, **kwargs):
        function_name = inspect.stack()[PREVIOUS_FUNCTION].function
        print("[ {time} ] [{level}] [{function}]: {message}".format(
            function=function_name,
            message=message,
            level=SystemMessage.yellow("warning"),
            time=time.strftime("%x %H:%M:%S %Z", time.localtime()),
            **kwargs
        ))

    @staticmethod
    def critical(message, **kwargs):
        function_name = inspect.stack()[PREVIOUS_FUNCTION].function
        print("[ {time} ] [{level}] [{function}]: {message}".format(
            function=function_name,
            message=message,
            level=SystemMessage.red_highlight("CRITICAL"),
            time=time.strftime("%x %H:%M:%S %Z", time.localtime()),
            **kwargs
        ))

    @staticmethod
    def red(message):
        return Style.BRIGHT + Fore.RED + message + Style.RESET_ALL

    @staticmethod
    def red_highlight(message):
        return Style.BRIGHT + Back.RED + Fore.WHITE + message + Style.RESET_ALL

    @staticmethod
    def blue(message):
        return Style.BRIGHT + Fore.BLUE + message + Style.RESET_ALL

    @staticmethod
    def blue_dim(message):
        return Style.DIM + Fore.BLUE + message + Style.RESET_ALL

    @staticmethod
    def green(message):
        return Style.BRIGHT + Fore.GREEN + message + Style.RESET_ALL

    @staticmethod
    def magenta(message):
        return Style.BRIGHT + Fore.MAGENTA + message + Style.RESET_ALL

    @staticmethod
    def magenta_dim(message):
        return Style.DIM + Fore.MAGENTA + message + Style.RESET_ALL

    @staticmethod
    def cyan(message):
        return Style.DIM + Fore.CYAN + message + Style.RESET_ALL

    @staticmethod
    def yellow(message):
        return Style.BRIGHT + Fore.YELLOW + message + Style.RESET_ALL

    @staticmethod
    def grey(message):
        return Style.BRIGHT + Fore.BLACK + message + Style.RESET_ALL



import click

from skriba.logger import setup_worker_logger


class WorkerLogger:
    def __init__(self, local_cache, log_params):
        self.logger = None
        self.worker = None
        self.local_cache = local_cache

        self.log_to_term = log_params['log_to_term']
        self.log_to_file = log_params['log_to_file']
        self.log_file = log_params['log_file']
        self.log_level = log_params['log_level']

    @property
    def get_logger(self):
        return self.logger

    def setup(self, worker):
        """
        Run when the plugin is attached to a worker. This happens when the plugin is registered
        and attached to existing workers, or when a worker is created after the plugin has been
        registered.
        """

        self.logger = setup_worker_logger(
            self.log_to_term,
            self.log_to_file,
            self.log_file,
            self.log_level,
            worker.id
        )

        self.logger.debug('Logger created on worker ' + str(worker.id) + ',*,' + str(worker.address))

        # Documentation https://distributed.dask.org/en/stable/worker.html#distributed.worker.Worker
        self.worker = worker

        if self.local_cache:
            ip = worker.address[worker.address.rfind('/') + 1:worker.address.rfind(':')]
            self.logger.debug(str(worker.id) + ',*,' + ip)
            worker.state.available_resources = {**worker.state.available_resources, **{ip: 1}}


# https://github.com/dask/distributed/issues/4169
@click.command()
@click.option("--local_cache", default=False)
# @click.option("--log_params", default={'log_to_term':True,'log_to_file':False,'log_file':'astrohack_',
# 'log_level':'DEBUG'})
@click.option("--log_to_term", default=True)
@click.option("--log_to_file", default=False)
@click.option("--log_file", default='astrohack_')
@click.option("--log_level", default='INFO')
async def dask_setup(worker, local_cache, log_to_term, log_to_file, log_file, log_level):
    log_params = {
        'log_to_term': log_to_term,
        'log_to_file': log_to_file,
        'log_file': log_file,
        'log_level': log_level
    }

    plugin = WorkerLogger(local_cache, log_params)

    await worker.client.register_worker_plugin(plugin, name='worker_logger')
